@@ -1,115 +1,100 @@
 use alloc::boxed::Box;
 use alloc::fmt;
 use alloc::vec::Vec;
+use derive_more::Display;
 use nom::branch::alt;
 use nom::bytes::streaming::tag;
 use nom::combinator::{map, opt};
 use nom::multi::separated_list1;
-use nom::sequence::{delimited, pair, preceded, tuple};
+use nom::sequence::{delimited, pair, preceded, terminated, tuple};
 #[cfg(feature = "serde")]
 use serde::Serialize;
 
 use crate::cnf;
 use crate::common::*;
 use crate::fof;
-use crate::utils::fmt_list;
+use crate::tfx;
+use crate::utils::Separated;
 use crate::{Error, Parse, Result};
 
 /// [`file_name`](http://tptp.org/TPTP/SyntaxBNF.html#file_name)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct FileName<'a>(pub SingleQuoted<'a>);
-impl_unit_anon_display! {FileName}
 
-parser! {
-    FileName,
-    map(SingleQuoted::parse, Self)
+impl<'a, E: Error<'a>> Parse<'a, E> for FileName<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(SingleQuoted::parse, Self)(x)
+    }
 }
 
 /// [`name_list`](http://tptp.org/TPTP/SyntaxBNF.html#name_list)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "{}", "Separated(',', _0)")]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct NameList<'a>(pub Vec<Name<'a>>);
 
-impl<'a> fmt::Display for NameList<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        fmt_list(f, ",", &self.0)
+impl<'a, E: Error<'a>> Parse<'a, E> for NameList<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            separated_list1(
+                delimited(ignored, tag(","), ignored),
+                Name::parse,
+            ),
+            Self,
+        )(x)
     }
 }
 
-parser! {
-    NameList,
-    map(
-        separated_list1(
-            tuple((ignored, tag(","), ignored)),
-            Name::parse,
-        ),
-        Self,
-    )
-}
-
 /// [`formula_role`](http://tptp.org/TPTP/SyntaxBNF.html#formula_role)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct FormulaRole<'a>(pub LowerWord<'a>);
-impl_unit_anon_display! {FormulaRole}
 
-parser! {
-    FormulaRole,
-    map(LowerWord::parse, FormulaRole)
-}
-
-/// [`formula_data`](http://tptp.org/TPTP/SyntaxBNF.html#formula_data)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
-#[cfg_attr(feature = "serde", derive(Serialize))]
-pub enum FormulaData<'a> {
-    Fof(fof::Formula<'a>),
-    Cnf(cnf::Formula<'a>),
-    Fot(fof::Term<'a>),
-}
-
-impl<'a> fmt::Display for FormulaData<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        use self::FormulaData::*;
-        match self {
-            Fof(fof) => write!(f, "$fof({})", fof),
-            Cnf(cnf) => write!(f, "$cnf({})", cnf),
-            Fot(cnf) => write!(f, "$fot({})", cnf),
-        }
+impl<'a, E: Error<'a>> Parse<'a, E> for FormulaRole<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(LowerWord::parse, FormulaRole)(x)
     }
 }
 
-parser! {
-    FormulaData,
-    preceded(
-        tag("$"),
-        alt((
-            map(
-                delimited(
-                    tuple((tag("fof"), ignored, tag("("), ignored)),
-                    fof::Formula::parse,
-                    tuple((ignored, tag(")"))),
+/// [`formula_data`](http://tptp.org/TPTP/SyntaxBNF.html#formula_data)
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[cfg_attr(feature = "serde", derive(Serialize))]
+pub enum FormulaData<'a> {
+    #[display(fmt = "$tff({})", _0)]
+    Tfx(tfx::Formula<'a>),
+    #[display(fmt = "$fof({})", _0)]
+    Fof(fof::Formula<'a>),
+    #[display(fmt = "$cnf({})", _0)]
+    Cnf(cnf::Formula<'a>),
+    #[display(fmt = "$fot({})", _0)]
+    Fot(fof::Term<'a>),
+}
+
+impl<'a, E: Error<'a>> Parse<'a, E> for FormulaData<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        preceded(
+            tag("$"),
+            alt((
+                preceded(
+                    tag("tff"),
+                    map(preceded(ignored, parens), Self::Tfx),
                 ),
-                Self::Fof,
-            ),
-            map(
-                delimited(
-                    tuple((tag("cnf"), ignored, tag("("), ignored)),
-                    cnf::Formula::parse,
-                    tuple((ignored, tag(")"))),
+                preceded(
+                    tag("fof"),
+                    map(preceded(ignored, parens), Self::Fof),
                 ),
-                Self::Cnf,
-            ),
-            map(
-                delimited(
-                    tuple((tag("fot"), ignored, tag("("), ignored)),
-                    fof::Term::parse,
-                    tuple((ignored, tag(")"))),
+                preceded(
+                    tag("cnf"),
+                    map(preceded(ignored, parens), Self::Cnf),
                 ),
-                Self::Fot,
-            ),
-        )),
-    )
+                preceded(
+                    tag("fot"),
+                    map(preceded(ignored, parens), Self::Fot),
+                ),
+            )),
+        )(x)
+    }
 }
 
 struct GeneralFunctionTail<'a>(GeneralTerms<'a>);
@@ -121,39 +106,35 @@ impl<'a> GeneralFunctionTail<'a> {
     }
 }
 
-parser! {
-    GeneralFunctionTail,
-    delimited(
-        delimited(ignored, tag("("), ignored),
-        map(GeneralTerms::parse, Self),
-        preceded(ignored, tag(")")),
-    )
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralFunctionTail<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(parens, Self)(x)
+    }
 }
 
 /// [`general_function`](http://tptp.org/TPTP/SyntaxBNF.html#general_function)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "{}({})", word, terms)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct GeneralFunction<'a> {
     pub word: AtomicWord<'a>,
     pub terms: GeneralTerms<'a>,
 }
 
-impl<'a> fmt::Display for GeneralFunction<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}({})", self.word, self.terms)
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralFunction<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            pair(
+                AtomicWord::parse,
+                preceded(ignored, GeneralFunctionTail::parse),
+            ),
+            |(word, tail)| tail.finish(word),
+        )(x)
     }
 }
 
-parser! {
-    GeneralFunction,
-    map(
-        pair(AtomicWord::parse, GeneralFunctionTail::parse),
-        |(word, tail)| tail.finish(word)
-    )
-}
-
 /// [`general_data`](http://tptp.org/TPTP/SyntaxBNF.html#general_data)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub enum GeneralData<'a> {
     Atomic(AtomicWord<'a>),
@@ -163,54 +144,47 @@ pub enum GeneralData<'a> {
     DistinctObject(DistinctObject<'a>),
     Formula(Box<FormulaData<'a>>),
 }
-impl_enum_anon_display!(
-    GeneralData,
-    Atomic,
-    Function,
-    Variable,
-    Number,
-    DistinctObject,
-    Formula
-);
 
-parser! {
-    GeneralData,
-    alt((
-        map(
-            pair(AtomicWord::parse, opt(GeneralFunctionTail::parse)),
-            |(word, tail)| if let Some(tail) = tail {
-                Self::Function(Box::new(tail.finish(word)))
-            } else {
-                Self::Atomic(word)
-            }
-        ),
-        map(Variable::parse, Self::Variable),
-        map(Number::parse, Self::Number),
-        map(DistinctObject::parse, Self::DistinctObject),
-        map(map(FormulaData::parse, Box::new), Self::Formula),
-    ))
-}
-
-/// [`general_terms`](http://tptp.org/TPTP/SyntaxBNF.html#general_terms)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
-#[cfg_attr(feature = "serde", derive(Serialize))]
-pub struct GeneralTerms<'a>(pub Vec<GeneralTerm<'a>>);
-
-impl<'a> fmt::Display for GeneralTerms<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        fmt_list(f, ",", &self.0)
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralData<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        alt((
+            map(
+                pair(
+                    AtomicWord::parse,
+                    opt(preceded(ignored, GeneralFunctionTail::parse)),
+                ),
+                |(word, tail)| {
+                    if let Some(tail) = tail {
+                        Self::Function(Box::new(tail.finish(word)))
+                    } else {
+                        Self::Atomic(word)
+                    }
+                },
+            ),
+            map(Variable::parse, Self::Variable),
+            map(Number::parse, Self::Number),
+            map(DistinctObject::parse, Self::DistinctObject),
+            map(map(FormulaData::parse, Box::new), Self::Formula),
+        ))(x)
     }
 }
 
-parser! {
-    GeneralTerms,
-    map(
-        separated_list1(
-            delimited(ignored, tag(","), ignored),
-            GeneralTerm::parse,
-        ),
-        Self,
-    )
+/// [`general_terms`](http://tptp.org/TPTP/SyntaxBNF.html#general_terms)
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "{}", "Separated(',', _0)")]
+#[cfg_attr(feature = "serde", derive(Serialize))]
+pub struct GeneralTerms<'a>(pub Vec<GeneralTerm<'a>>);
+
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralTerms<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            separated_list1(
+                delimited(ignored, tag(","), ignored),
+                GeneralTerm::parse,
+            ),
+            Self,
+        )(x)
+    }
 }
 
 /// [`general_list`](http://tptp.org/TPTP/SyntaxBNF.html#general_list)
@@ -227,81 +201,73 @@ impl<'a> fmt::Display for GeneralList<'a> {
     }
 }
 
-parser! {
-    GeneralList,
-    map(
-        delimited(
-            pair(tag("["), ignored),
-            opt(GeneralTerms::parse),
-            pair(ignored, tag("]")),
-        ),
-        Self,
-    )
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralList<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            delimited(
+                tag("["),
+                delimited(ignored, opt(GeneralTerms::parse), ignored),
+                tag("]"),
+            ),
+            Self,
+        )(x)
+    }
 }
 
 /// [`general_term`](http://tptp.org/TPTP/SyntaxBNF.html#general_term)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub enum GeneralTerm<'a> {
     Data(GeneralData<'a>),
+    #[display(fmt = "{}:{}", _0, _1)]
     Colon(GeneralData<'a>, Box<GeneralTerm<'a>>),
     List(GeneralList<'a>),
 }
 
-impl<'a> fmt::Display for GeneralTerm<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        use self::GeneralTerm::*;
-        match self {
-            Data(d) => write!(f, "{}", d),
-            Colon(d, t) => write!(f, "{}:{}", d, t),
-            List(l) => write!(f, "{}", l),
-        }
+impl<'a, E: Error<'a>> Parse<'a, E> for GeneralTerm<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        alt((
+            map(GeneralList::parse, Self::List),
+            map(
+                pair(
+                    GeneralData::parse,
+                    opt(preceded(
+                        pair(ignored, tag(":")),
+                        preceded(ignored, Self::parse),
+                    )),
+                ),
+                |(left, right)| {
+                    if let Some(right) = right {
+                        Self::Colon(left, Box::new(right))
+                    } else {
+                        Self::Data(left)
+                    }
+                },
+            ),
+        ))(x)
     }
 }
 
-parser! {
-    GeneralTerm,
-    alt((
-        map(GeneralList::parse, Self::List),
-        map(
-            pair(
-                GeneralData::parse,
-                opt(preceded(
-                    delimited(ignored, tag(":"), ignored),
-                    Self::parse,
-                )),
-            ),
-            |(left, right)| {
-                if let Some(right) = right {
-                    Self::Colon(left, Box::new(right))
-                } else {
-                    Self::Data(left)
-                }
-            },
-        ),
-    ))
-}
-
 /// [`source`](http://tptp.org/TPTP/SyntaxBNF.html#source)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct Source<'a>(pub GeneralTerm<'a>);
-impl_unit_anon_display! {Source}
 
-parser! {
-    Source,
-    map(GeneralTerm::parse, Self)
+impl<'a, E: Error<'a>> Parse<'a, E> for Source<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(GeneralTerm::parse, Self)(x)
+    }
 }
 
 /// [`useful_info`](http://tptp.org/TPTP/SyntaxBNF.html#useful_info)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct UsefulInfo<'a>(pub GeneralList<'a>);
-impl_unit_anon_display! {UsefulInfo}
 
-parser! {
-    UsefulInfo,
-    map(GeneralList::parse, Self)
+impl<'a, E: Error<'a>> Parse<'a, E> for UsefulInfo<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(GeneralList::parse, Self)(x)
+    }
 }
 
 /// [`optional_info`](http://tptp.org/TPTP/SyntaxBNF.html#optional_info)
@@ -318,12 +284,13 @@ impl<'a> fmt::Display for OptionalInfo<'a> {
     }
 }
 
-parser! {
-    OptionalInfo,
-    map(
-        opt(preceded(pair(tag(","), ignored), UsefulInfo::parse)),
-        Self,
-    )
+impl<'a, E: Error<'a>> Parse<'a, E> for OptionalInfo<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            opt(preceded(tag(","), preceded(ignored, UsefulInfo::parse))),
+            Self,
+        )(x)
+    }
 }
 
 /// [`annotations`](http://tptp.org/TPTP/SyntaxBNF.html#annotations)
@@ -340,18 +307,25 @@ impl<'a> fmt::Display for Annotations<'a> {
     }
 }
 
-parser! {
-    Annotations,
-    map(
-        opt(preceded(
-            pair(tag(","), ignored),
-            map(
-                pair(Source::parse, preceded(ignored, OptionalInfo::parse)),
-                Box::new
-            )
-        )),
-        Self
-    )
+impl<'a, E: Error<'a>> Parse<'a, E> for Annotations<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            opt(preceded(
+                pair(tag(","), ignored),
+                map(
+                    pair(
+                        Source::parse,
+                        map(
+                            opt(preceded(ignored, OptionalInfo::parse)),
+                            |info| info.unwrap_or(OptionalInfo(None)),
+                        ),
+                    ),
+                    Box::new,
+                ),
+            )),
+            Self,
+        )(x)
+    }
 }
 
 /// [`formula_selection`](http://tptp.org/TPTP/SyntaxBNF.html#formula_selection)
@@ -368,49 +342,60 @@ impl<'a> fmt::Display for FormulaSelection<'a> {
     }
 }
 
-parser! {
-    FormulaSelection,
-    map(
-        opt(delimited(
-            tuple((tag(","), ignored, tag("["), ignored)),
-            NameList::parse,
-            tuple((ignored, tag("]"))),
-        )),
-        Self,
-    )
+impl<'a, E: Error<'a>> Parse<'a, E> for FormulaSelection<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            opt(preceded(tag(","), delimited(ignored, brackets, ignored))),
+            Self,
+        )(x)
+    }
 }
 
 /// [`include`](http://tptp.org/TPTP/SyntaxBNF.html#include)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "include({}{}).", file_name, selection)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct Include<'a> {
     pub file_name: FileName<'a>,
     pub selection: FormulaSelection<'a>,
 }
 
-impl<'a> fmt::Display for Include<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "include({}{}).", self.file_name, self.selection)
+impl<'a, E: Error<'a>> Parse<'a, E> for Include<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        map(
+            preceded(
+                tag("include"),
+                terminated(
+                    delimited(
+                        ignored,
+                        delimited(
+                            tag("("),
+                            delimited(
+                                ignored,
+                                pair(
+                                    FileName::parse,
+                                    preceded(ignored, FormulaSelection::parse),
+                                ),
+                                ignored,
+                            ),
+                            tag(")"),
+                        ),
+                        ignored,
+                    ),
+                    tag("."),
+                ),
+            ),
+            |(file_name, selection)| Self {
+                file_name,
+                selection,
+            },
+        )(x)
     }
 }
 
-parser! {
-    Include,
-    map(
-        delimited(
-            tuple((tag("include"), ignored, tag("("), ignored)),
-            pair(FileName::parse, preceded(ignored, FormulaSelection::parse)),
-            tuple((ignored, tag(")"), ignored, tag("."))),
-        ),
-        |(file_name, selection)| Self {
-            file_name,
-            selection,
-        },
-    )
-}
-
 /// helper struct to share common fields - thanks to Michael Färber
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "({},{},{}{}).", name, role, formula, annotations)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct Annotated<'a, T> {
     pub name: Name<'a>,
@@ -419,34 +404,30 @@ pub struct Annotated<'a, T> {
     pub annotations: Annotations<'a>,
 }
 
-impl<'a, T: fmt::Display> fmt::Display for Annotated<'a, T> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(
-            f,
-            "({},{},{}{}).",
-            self.name, self.role, self.formula, self.annotations
-        )
-    }
-}
-
 impl<'a, E: Error<'a>, T: Parse<'a, E>> Parse<'a, E> for Annotated<'a, T> {
     fn parse(x: &'a [u8]) -> Result<'a, Self, E> {
         map(
-            delimited(
-                pair(tag("("), ignored),
-                tuple((
-                    Name::parse,
-                    preceded(
-                        delimited(ignored, tag(","), ignored),
-                        FormulaRole::parse,
+            preceded(
+                tag("("),
+                terminated(
+                    delimited(
+                        ignored,
+                        tuple((
+                            Name::parse,
+                            preceded(
+                                delimited(ignored, tag(","), ignored),
+                                FormulaRole::parse,
+                            ),
+                            preceded(
+                                delimited(ignored, tag(","), ignored),
+                                map(T::parse, Box::new),
+                            ),
+                            preceded(ignored, Annotations::parse),
+                        )),
+                        ignored,
                     ),
-                    preceded(
-                        delimited(ignored, tag(","), ignored),
-                        map(T::parse, Box::new),
-                    ),
-                    preceded(ignored, Annotations::parse),
-                )),
-                pair(ignored, tag(")")),
+                    tuple((tag(")"), ignored, tag("."))),
+                ),
             ),
             |(name, role, formula, annotations)| Self {
                 name,
@@ -458,78 +439,76 @@ impl<'a, E: Error<'a>, T: Parse<'a, E>> Parse<'a, E> for Annotated<'a, T> {
     }
 }
 
+/// [`tff_annotated`](http://tptp.org/TPTP/SyntaxBNF.html#tff_annotated)
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "tff{}", _0)]
+#[cfg_attr(feature = "serde", derive(Serialize))]
+pub struct TfxAnnotated<'a>(pub Annotated<'a, tfx::Formula<'a>>);
+
+impl<'a, E: Error<'a>> Parse<'a, E> for TfxAnnotated<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        preceded(tag("tff"), preceded(ignored, map(Annotated::parse, Self)))(x)
+    }
+}
+
 /// [`fof_annotated`](http://tptp.org/TPTP/SyntaxBNF.html#fof_annotated)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "fof{}", _0)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct FofAnnotated<'a>(pub Annotated<'a, fof::Formula<'a>>);
 
-impl<'a> fmt::Display for FofAnnotated<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "fof{}", self.0)
+impl<'a, E: Error<'a>> Parse<'a, E> for FofAnnotated<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        preceded(tag("fof"), preceded(ignored, map(Annotated::parse, Self)))(x)
     }
 }
 
-parser! {
-    FofAnnotated,
-    delimited(
-        pair(tag("fof"), ignored),
-        map(Annotated::parse, Self),
-        pair(ignored, tag(".")),
-    )
-}
-
 /// [`cnf_annotated`](http://tptp.org/TPTP/SyntaxBNF.html#cnf_annotated)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[display(fmt = "cnf{}", _0)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub struct CnfAnnotated<'a>(pub Annotated<'a, cnf::Formula<'a>>);
 
-impl<'a> fmt::Display for CnfAnnotated<'a> {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "cnf{}", self.0)
+impl<'a, E: Error<'a>> Parse<'a, E> for CnfAnnotated<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        preceded(tag("cnf"), preceded(ignored, map(Annotated::parse, Self)))(x)
     }
 }
 
-parser! {
-    CnfAnnotated,
-    delimited(
-        pair(tag("cnf"), ignored),
-        map(Annotated::parse, Self),
-        pair(ignored, tag(".")),
-    )
-}
-
 /// [`annotated_formula`](http://tptp.org/TPTP/SyntaxBNF.html#annotated_formula)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub enum AnnotatedFormula<'a> {
+    Tfx(Box<TfxAnnotated<'a>>),
     Fof(Box<FofAnnotated<'a>>),
     Cnf(Box<CnfAnnotated<'a>>),
 }
-impl_enum_anon_display! {AnnotatedFormula, Fof, Cnf}
 
-parser! {
-    AnnotatedFormula,
-    alt((
-        map(map(FofAnnotated::parse, Box::new), Self::Fof),
-        map(map(CnfAnnotated::parse, Box::new), Self::Cnf),
-    ))
+impl<'a, E: Error<'a>> Parse<'a, E> for AnnotatedFormula<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        alt((
+            map(map(TfxAnnotated::parse, Box::new), Self::Tfx),
+            map(map(FofAnnotated::parse, Box::new), Self::Fof),
+            map(map(CnfAnnotated::parse, Box::new), Self::Cnf),
+        ))(x)
+    }
 }
 
 /// [`TPTP_input`](http://tptp.org/TPTP/SyntaxBNF.html#TPTP_input)
-#[derive(Clone, Debug, PartialOrd, Ord, PartialEq, Eq, Hash)]
+#[derive(Clone, Debug, Display, PartialOrd, Ord, PartialEq, Eq, Hash)]
 #[cfg_attr(feature = "serde", derive(Serialize))]
 pub enum TPTPInput<'a> {
     Annotated(Box<AnnotatedFormula<'a>>),
     Include(Box<Include<'a>>),
 }
-impl_enum_anon_display! {TPTPInput, Annotated, Include}
 
-parser! {
-    TPTPInput,
-    alt((
-        map(map(AnnotatedFormula::parse, Box::new), Self::Annotated),
-        map(map(Include::parse, Box::new), Self::Include),
-    ))
+impl<'a, E: Error<'a>> Parse<'a, E> for TPTPInput<'a> {
+    fn parse(x: &'a [u8]) -> Result<Self, E> {
+        alt((
+            map(map(AnnotatedFormula::parse, Box::new), Self::Annotated),
+            map(map(Include::parse, Box::new), Self::Include),
+        ))(x)
+    }
 }
 
 #[cfg(test)]
@@ -567,6 +546,7 @@ mod tests {
     #[test]
     fn test_formula_data() {
         check_size::<FormulaData>();
+        parse::<FormulaData>(b"$tff ( p )\0");
         parse::<FormulaData>(b"$fof ( p )\0");
         parse::<FormulaData>(b"$cnf ( p )\0");
         parse::<FormulaData>(b"$fot ( t )\0");
@@ -612,6 +592,16 @@ mod tests {
         parse::<Annotations>(b", c , [X]\0");
     }
 
+    #[test]
+    fn test_tfx_annotated() {
+        check_size::<TfxAnnotated>();
+        parse::<TfxAnnotated>(b"tff ( test , axiom , $true ) .\0");
+        parse::<TfxAnnotated>(b"tff ( test , axiom , $true , unknown ) .\0");
+        parse::<TfxAnnotated>(
+            b"tff ( test , axiom , $true , unknown , [] ) .\0",
+        );
+    }
+
     #[test]
     fn test_fof_annotated() {
         check_size::<FofAnnotated>();
@@ -635,6 +625,7 @@ mod tests {
     #[test]
     fn test_annotated_formula() {
         check_size::<AnnotatedFormula>();
+        parse::<AnnotatedFormula>(b"tff ( test , axiom , $true ) .\0");
         parse::<AnnotatedFormula>(b"fof ( test , axiom , $true ) .\0");
         parse::<AnnotatedFormula>(b"cnf ( test , axiom , $true ) .\0");
     }
@@ -670,6 +661,7 @@ mod tests {
     fn test_tptp_input() {
         check_size::<TPTPInput>();
         parse::<TPTPInput>(b"include ( 'test' ) .\0");
+        parse::<TPTPInput>(b"tff ( test , axiom , $true ) .\0");
         parse::<TPTPInput>(b"fof ( test , axiom , $true ) .\0");
         parse::<TPTPInput>(b"cnf ( test , axiom , $true ) .\0");
     }
